/*---------------------------------------------------------------------------*\
 *           interactive networked Virtual Reality system (inVRs)            *
 *                                                                           *
 *    Copyright (C) 2005-2009 by the Johannes Kepler University, Linz        *
 *                                                                           *
 *                            www.inVRs.org                                  *
 *                                                                           *
 *              contact: canthes@inVRs.org, rlander@inVRs.org                *
\*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*\
 *                                License                                    *
 *                                                                           *
 * This library is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU Library General Public License as published    *
 * by the Free Software Foundation, version 2.                               *
 *                                                                           *
 * This library is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 * Library General Public License for more details.                          *
 *                                                                           *
 * You should have received a copy of the GNU Library General Public         *
 * License along with this library; if not, write to the Free Software       *
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 *
\*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*\
 *                           Project: Avatara                                *
 *                                                                           *
 * The Avatara library was developed during a practical at the Johannes      *
 * Kepler University, Linz in 2005 by Helmut Garstenauer                     *
 * (helmut@digitalrune.com) and Martin Garstenauer (martin@digitalrune.com)  *
 *                                                                           *
\*---------------------------------------------------------------------------*/

//_____________________________________________________________
//
/// \file OSGAvatar.cpp
/// Implementation of avatar for OpenSG.
//_____________________________________________________________

#include "OSGAvatar.h"
#include "Debug.h"
#include <OpenSG/OSGSimpleGeometry.h>
#include <OpenSG/OSGBaseFunctions.h>
#include <OpenSG/OSGSimpleMaterial.h>
#include <OpenSG/OSGSimpleTexturedMaterial.h>

OSG_USING_NAMESPACE;
using namespace Algebra;


namespace Avatara
{

  //-------------------------------------------------------
  /// Initializes empty avatar object.
  //-------------------------------------------------------
  OSGAvatar::OSGAvatar()
  {
  }


  //-------------------------------------------------------
  /// Cleans up resources.
  //-------------------------------------------------------
  OSGAvatar::~OSGAvatar()
  {
  }


  //-------------------------------------------------------
  /// Builds an OpenSG geometry of the avatar.
  ///
  /// \param image IN: Avatar texture.
  /// \return OpenSG geometry.
  //-------------------------------------------------------
  OSG::GeometryPtr OSGAvatar::MakeAvatarGeo(OSG::ImagePtr image)
  {
    ASSERT(pModel != NULL);

    //
    // ----- type and length attributes
    //
    // Avatar consists of triangles only.
    // 3 vertices per face (triangle).
    GeoPTypesPtr type = GeoPTypesUI8::create();
    GeoPLengthsPtr lens = GeoPLengthsUI32::create();
    beginEditCP(type, GeoPTypesUI8::GeoPropDataFieldMask);
    beginEditCP(lens, GeoPLengthsUI32::GeoPropDataFieldMask);
    {
      type->push_back(GL_TRIANGLES);
      lens->push_back(3*pModel->GetNoOfFaces());
    }
    endEditCP  (lens, GeoPLengthsUI32::GeoPropDataFieldMask);
    endEditCP  (type, GeoPTypesUI8::GeoPropDataFieldMask);

    //
    // ----- vertex positions
    //
    // One position per vertex is stored.
    GeoPositions3fPtr pnts = GeoPositions3f::create();
    beginEditCP(pnts, GeoPositions3f::GeoPropDataFieldMask);
    {
      int i=0;
      for(i=0; i<pModel->GetNoOfVertices(); i++)
      {
        Algebra::Vector pnt = pModel->GetVertex(i).Position();
        pnts->push_back(Pnt3f(pnt.X(), pnt.Y(), pnt.Z()));
      }
    }
    endEditCP  (pnts, GeoPositions3f::GeoPropDataFieldMask);

    //
    // ----- color, normal vectors, texture coordinates
    //
    GeoColors3fPtr colors = GeoColors3f::create();
    GeoNormals3fPtr norms = GeoNormals3f::create();
    GeoTexCoords2fPtr tex = GeoTexCoords2f::create();
    beginEditCP(colors, GeoColors3f::GeoPropDataFieldMask);
    beginEditCP(norms, GeoNormals3f::GeoPropDataFieldMask);
    beginEditCP(tex, GeoTexCoords2f::GeoPropDataFieldMask);
    {
      int i=0;
      for(i=0; i<pModel->GetNoOfFaces(); i++)
      {
        int j=0;
        Face face = pModel->GetFace(i);
        for (j=0; j<3; j++)
        {
          Color col = face.pColors[j];
          colors->push_back(Color3f(col.r, col.g, col.b));

          Algebra::Vector normal;
          if(face.shading == Smooth)
            normal = pModel->GetVertex(face.pVertexIndices[j]).Normal();
          else
            normal = face.normal;
          normal.Normalize();
          norms->push_back(Vec3f(normal.X(), normal.Y(), normal.Z()));

          tex->push_back(Vec2f(face.pTextureCoords[j].u, face.pTextureCoords[j].v));
        }
      }
    }
    endEditCP  (tex, GeoTexCoords2f::GeoPropDataFieldMask);
    endEditCP  (norms, GeoNormals3f::GeoPropDataFieldMask);
    endEditCP  (colors, GeoPositions3f::GeoPropDataFieldMask);

    //
    // ----- index list
    //
    // 2 indices per vertex. First index is for positions. Second index is for
    // color, normal vectors and texture coords.
    // 3 vertices per face.
    GeoIndicesUI32Ptr indices = GeoIndicesUI32::create();
    beginEditCP(indices, GeoIndicesUI32::GeoPropDataFieldMask);
    {
      int i=0;
      for (i=0; i<pModel->GetNoOfFaces(); i++)
      {
        Face face = pModel->GetFace(i);
        int j=0;
        for (j=0; j<3; j++)
        {
          indices->push_back(face.pVertexIndices[j]); // position index
          indices->push_back(i*3+j);                  // color, normal, texture index
        }
      }
    }
    endEditCP  (indices, GeoIndicesUI32::GeoPropDataFieldMask);

    //
    // ----- Textured material
    //
    SimpleTexturedMaterialPtr mat = SimpleTexturedMaterial::create();
    beginEditCP(mat);
    {
      mat->setAmbient      (Color3f(0.3,0.3,0.3));
      mat->setDiffuse      (Color3f(0.8,0.8,0.8));
      mat->setEmission     (Color3f(0.0,0.0,0.0));
      mat->setSpecular     (Color3f(0.8,0.8,0.8));
      mat->setShininess    (20);
      mat->setTransparency (0);
      mat->setColorMaterial(GL_NONE);
      mat->setImage        (image);
      mat->setMinFilter    (GL_LINEAR);
      mat->setMagFilter    (GL_LINEAR);
      mat->setEnvMode      (GL_MODULATE);
      mat->setEnvMap       (false);
    }
    endEditCP  (mat);

    //
    // ----- create geometry
    //
    geo = Geometry::create();
    beginEditCP(geo);
    {
      geo->setTypes    (type);
      geo->setLengths  (lens);
      geo->setPositions(pnts);
      geo->setColors   (colors);
      geo->setNormals  (norms);
      geo->setTexCoords(tex);
      geo->setIndices  (indices);
      geo->getIndexMapping().push_back(Geometry::MapPosition);
      geo->getIndexMapping().push_back(Geometry::MapColor  |
                                       Geometry::MapNormal |
                                       Geometry::MapTexCoords);
      //geo->setMaterial (getDefaultMaterial());
      geo->setMaterial (mat);

      // turn of display list caching
      geo->setDlistCache(false);
    }
    endEditCP  (geo);

    return geo;
  }


  //-------------------------------------------------------
  /// Updates avatar geometry to a new pose.
  //-------------------------------------------------------
  void OSGAvatar::UpdateAvatarGeo()
  {
    // get position vector from the geometry
    GeoPositionsPtr pos = geo->getPositions();

    ASSERT(pModel->GetNoOfVertices() == pos->getSize());

    // Calculate the new pose of the animation
    CalculatePose();

    // discard the old positions and add the new skinned positions
    beginEditCP(pos);
    for(UInt32 i = 0; i < pos->getSize(); i++)
    {
      Algebra::Vector pnt = SkinVertex(pModel->GetVertex(i));
      Pnt3f p(pnt.X(), pnt.Y(), pnt.Z());
      pos->setValue(p, i);
    }
    endEditCP  (pos);

    // right now the geometry doesn't notice changes to the properties, it has
    // to be notified explicitly
    beginEditCP(geo, Geometry::PositionsFieldMask);
    endEditCP  (geo, Geometry::PositionsFieldMask);
  }
}

